"use strict";var EventEmitter = require("events-light");
var RenderResult = require("../RenderResult");
var attrsHelper = require("./helpers/attrs");
var morphdom = require("./morphdom");
var vdom = require("./vdom");
var VElement = vdom.bz_;
var VDocumentFragment = vdom.bA_;
var VText = vdom.bB_;
var VComponent = vdom.bC_;
var VFragment = vdom.bD_;
var virtualizeHTML = vdom.bE_;

var EVENT_UPDATE = "update";
var EVENT_FINISH = "finish";

function State(tree) {
  this.bF_ = new EventEmitter();
  this.bG_ = tree;
  this.bH_ = false;
}

function AsyncVDOMBuilder(globalData, parentNode, parentOut) {
  if (!parentNode) {
    parentNode = new VDocumentFragment();
  }

  var state;

  if (parentOut) {
    state = parentOut.z_;
  } else {
    state = new State(parentNode);
  }

  this.bI_ = 1;
  this.bJ_ = 0;
  this.bK_ = null;
  this.bL_ = parentOut;

  this.data = {};
  this.z_ = state;
  this.ab_ = parentNode;
  this.global = globalData || {};
  this.bM_ = [parentNode];
  this.bN_ = false;
  this.bO_ = undefined;
  this.b_ = null;

  this.a__ = null;
  this.aa_ = null;
  this.bc_ = null;
}

var proto = AsyncVDOMBuilder.prototype = {
  bt_: true,
  B_: typeof document === "object" && document,

  bc: function (component, key, ownerComponent) {
    var vComponent = new VComponent(component, key, ownerComponent);
    return this.bP_(vComponent, 0, true);
  },

  bd_: function (component, key, ownerComponent) {
    var vComponent = new VComponent(component, key, ownerComponent, true);
    this.bP_(vComponent, 0);
  },

  bP_: function (child, childCount, pushToStack) {
    this.ab_.bQ_(child);
    if (pushToStack === true) {
      this.bM_.push(child);
      this.ab_ = child;
    }
    return childCount === 0 ? this : child;
  },

  element: function (tagName, attrs, key, component, childCount, flags, props) {
    var element = new VElement(
      tagName,
      attrs,
      key,
      component,
      childCount,
      flags,
      props
    );
    return this.bP_(element, childCount);
  },

  bm_: function (tagName, attrs, key, componentDef, props) {
    return this.element(
      tagName,
      attrsHelper(attrs),
      key,
      componentDef.s_,
      0,
      0,
      props
    );
  },

  n: function (node, component) {
    // NOTE: We do a shallow clone since we assume the node is being reused
    //       and a node can only have one parent node.
    var clone = node.bR_();
    this.node(clone);
    clone._M_ = component;

    return this;
  },

  node: function (node) {
    this.ab_.bQ_(node);
    return this;
  },

  text: function (text, ownerComponent) {
    var type = typeof text;

    if (type != "string") {
      if (text == null) {
        return;
      } else if (type === "object") {
        if (text.toHTML) {
          return this.h(text.toHTML(), ownerComponent);
        }
      }

      text = text.toString();
    }

    this.ab_.bQ_(new VText(text, ownerComponent));
    return this;
  },

  html: function (html, ownerComponent) {
    if (html != null) {
      var vdomNode = virtualizeHTML(html, ownerComponent);
      this.node(vdomNode);
    }

    return this;
  },

  beginElement: function (
  tagName,
  attrs,
  key,
  component,
  childCount,
  flags,
  props)
  {
    var element = new VElement(
      tagName,
      attrs,
      key,
      component,
      childCount,
      flags,
      props
    );
    this.bP_(element, childCount, true);
    return this;
  },

  bk_: function (tagName, attrs, key, componentDef, props) {
    return this.beginElement(
      tagName,
      attrsHelper(attrs),
      key,
      componentDef.s_,
      0,
      0,
      props
    );
  },

  bf: function (key, component, preserve) {
    var fragment = new VFragment(key, component, preserve);
    this.bP_(fragment, null, true);
    return this;
  },

  ef: function () {
    this.endElement();
  },

  endElement: function () {
    var stack = this.bM_;
    stack.pop();
    this.ab_ = stack[stack.length - 1];
  },

  end: function () {
    this.ab_ = undefined;

    var remaining = --this.bI_;
    var parentOut = this.bL_;

    if (remaining === 0) {
      if (parentOut) {
        parentOut.bS_();
      } else {
        this.bT_();
      }
    } else if (remaining - this.bJ_ === 0) {
      this.bU_();
    }

    return this;
  },

  bS_: function () {
    var remaining = --this.bI_;

    if (remaining === 0) {
      var parentOut = this.bL_;
      if (parentOut) {
        parentOut.bS_();
      } else {
        this.bT_();
      }
    } else if (remaining - this.bJ_ === 0) {
      this.bU_();
    }
  },

  bT_: function () {
    var state = this.z_;
    state.bH_ = true;
    state.bF_.emit(EVENT_FINISH, this.bv_());
  },

  bU_: function () {
    var lastArray = this._last;

    var i = 0;

    function next() {
      if (i === lastArray.length) {
        return;
      }
      var lastCallback = lastArray[i++];
      lastCallback(next);

      if (!lastCallback.length) {
        next();
      }
    }

    next();
  },

  error: function (e) {
    try {
      this.emit("error", e);
    } finally {
      // If there is no listener for the error event then it will
      // throw a new Error here. In order to ensure that the async fragment
      // is still properly ended we need to put the end() in a `finally`
      // block
      this.end();
    }

    return this;
  },

  beginAsync: function (options) {
    if (this.bN_) {
      throw Error(
        "Tried to render async while in sync mode. Note: Client side await is not currently supported in re-renders (Issue: #942)."
      );
    }

    var state = this.z_;

    if (options) {
      if (options.last) {
        this.bJ_++;
      }
    }

    this.bI_++;

    var documentFragment = this.ab_.bV_();
    var asyncOut = new AsyncVDOMBuilder(this.global, documentFragment, this);

    state.bF_.emit("beginAsync", {
      out: asyncOut,
      parentOut: this
    });

    return asyncOut;
  },

  createOut: function () {
    return new AsyncVDOMBuilder(this.global);
  },

  flush: function () {
    var events = this.z_.bF_;

    if (events.listenerCount(EVENT_UPDATE)) {
      events.emit(EVENT_UPDATE, new RenderResult(this));
    }
  },

  ah_: function () {
    return this.z_.bG_;
  },

  bv_: function () {
    return this.bW_ || (this.bW_ = new RenderResult(this));
  },

  on: function (event, callback) {
    var state = this.z_;

    if (event === EVENT_FINISH && state.bH_) {
      callback(this.bv_());
    } else if (event === "last") {
      this.onLast(callback);
    } else {
      state.bF_.on(event, callback);
    }

    return this;
  },

  once: function (event, callback) {
    var state = this.z_;

    if (event === EVENT_FINISH && state.bH_) {
      callback(this.bv_());
    } else if (event === "last") {
      this.onLast(callback);
    } else {
      state.bF_.once(event, callback);
    }

    return this;
  },

  emit: function (type, arg) {
    var events = this.z_.bF_;
    switch (arguments.length) {
      case 1:
        events.emit(type);
        break;
      case 2:
        events.emit(type, arg);
        break;
      default:
        events.emit.apply(events, arguments);
        break;
    }
    return this;
  },

  removeListener: function () {
    var events = this.z_.bF_;
    events.removeListener.apply(events, arguments);
    return this;
  },

  sync: function () {
    this.bN_ = true;
  },

  isSync: function () {
    return this.bN_;
  },

  onLast: function (callback) {
    var lastArray = this._last;

    if (lastArray === undefined) {
      this._last = [callback];
    } else {
      lastArray.push(callback);
    }

    return this;
  },

  ag_: function (host) {
    var node = this.bO_;
    if (!node) {
      var vdomTree = this.ah_();

      if (!host) host = this.B_;
      this.bO_ = node = vdomTree.bo_(host, null);
      morphdom(node, vdomTree, host, this.b_);
    }
    return node;
  },

  toString: function (host) {
    var docFragment = this.ag_(host);
    var html = "";

    var child = docFragment.firstChild;
    while (child) {
      var nextSibling = child.nextSibling;
      if (child.nodeType != 1) {
        var container = docFragment.ownerDocument.createElement("div");
        container.appendChild(child.cloneNode());
        html += container.innerHTML;
      } else {
        html += child.outerHTML;
      }

      child = nextSibling;
    }

    return html;
  },

  then: function (fn, fnErr) {
    // eslint-disable-next-line @typescript-eslint/no-this-alias
    var out = this;
    var promise = new Promise(function (resolve, reject) {
      out.on("error", reject).on(EVENT_FINISH, function (result) {
        resolve(result);
      });
    });

    return Promise.resolve(promise).then(fn, fnErr);
  },

  catch: function (fnErr) {
    return this.then(undefined, fnErr);
  },

  isVDOM: true,

  c: function (componentDef, key, customEvents) {
    this.a__ = componentDef;
    this.aa_ = key;
    this.bc_ = customEvents;
  }
};

proto.e = proto.element;
proto.be = proto.beginElement;
proto.ee = proto.bl_ = proto.endElement;
proto.t = proto.text;
proto.h = proto.w = proto.write = proto.html;

module.exports = AsyncVDOMBuilder;